from __future__ import annotations

import typing as t

import numpy as np
import pyqtgraph as pg
from pyqtgraph.Qt import QtWidgets
from utilitys.processing import *

__all__ = ['ProcessIO', 'ProcessStage', 'NestedProcess', 'ImageProcess',
           'AtomicProcess', 'GlobalPredictionProcess']

_infoType = t.List[t.Union[t.List, t.Dict[str, t.Any]]]
StrList = t.List[str]
StrCol = t.Collection[str]

class ImageProcess(NestedProcess):
  mainInputKeys = ['image']
  mainResultKeys = ['image']

  @classmethod
  def _cmpPrevCurInfos(cls, prevInfos: t.List[dict], infos: t.List[dict]):
    validInfos = super()._cmpPrevCurInfos(prevInfos, infos)
    # Iterate backwards to facilitate entry deletion
    for ii in range(len(validInfos)-1,-1,-1):
      info = validInfos[ii]
      duplicateKeys = {'name'}
      for k, v in info.items():
        if v is cls._DUPLICATE_INFO:
          duplicateKeys.add(k)
      if len(info.keys() - duplicateKeys) == 0:
        del validInfos[ii]
    return validInfos


  def _stageSummaryWidget(self):
    infoToDisplay = self.getStageInfos()

    numStages = len(infoToDisplay)
    nrows = np.sqrt(numStages).astype(int)
    ncols = np.ceil(numStages/nrows)
    outGrid = pg.GraphicsLayoutWidget()
    sizeToAxMapping: t.Dict[tuple, pg.PlotItem] = {}
    for ii, info in enumerate(infoToDisplay):
      pltItem: pg.PlotItem = outGrid.addPlot(title=info.get('name', None))
      pltItem.getViewBox().invertY(True)
      npImg = info['image']
      sameSizePlt = sizeToAxMapping.get(npImg.shape[:2], None)
      if sameSizePlt is not None:
        pltItem.setXLink(sameSizePlt)
        pltItem.setYLink(sameSizePlt)
      sizeToAxMapping[npImg.shape[:2]] = pltItem
      imgItem = pg.ImageItem(npImg)
      pltItem.addItem(imgItem)

      if ii % ncols == ncols-1:
        outGrid.nextRow()
    # See https://github.com/pyqtgraph/pyqtgraph/issues/1348. strange zooming occurs
    # if aspect is locked on all figures
    for ax in sizeToAxMapping.values():
      ax.getViewBox().setAspectLocked(True)
    oldClose = outGrid.closeEvent
    def newClose(ev):
      del _winRefs[outGrid]
      oldClose(ev)
    # Windows that go out of scope get garbage collected. Prevent that here
    _winRefs[outGrid] = outGrid
    outGrid.closeEvent = newClose

    return outGrid

  def getStageInfos(self, ignoreDuplicates=True):
    infos = super().getStageInfos(ignoreDuplicates)
    # Add entry for initial image since it will be missed when just searching for
    # stage outputs
    infos.insert(0, {'name': 'Initial Image', 'image': self.input['image']})
    return infos

_winRefs = {}

class GlobalPredictionProcess(ImageProcess):
  def _stageSummaryWidget(self):
    return QtWidgets.QWidget()
  mainResultKeys = ['components']

class CategoricalProcess(NestedProcess):
  def _stageSummaryWidget(self):
    pass

  mainResultKeys = ['categories', 'confidences']


# from .annotator_ui import *
# from .clickables import *
# from .drawopts import *
# from .imageareas import *
# from .parameditors import *
# from .procwrapper import *
# from .regions import *
# from .rois import *
# from .tableview import *



from pathlib import Path
from typing import List, Union
from warnings import warn

from pyqtgraph.Qt import QtCore, QtWidgets
from pyqtgraph.parametertree.parameterTypes import GroupParameter, Parameter
from utilitys import ParamEditor, ParamEditorDockGrouping, fns, widgets as uw
from utilitys.fns import warnLater
from utilitys.params.pgregistered import ShortcutKeySeqParameter as ShcKeySeq


from s3a.constants import QUICK_LOAD_DIR
from ..generalutils import lower_NoSpaces

class EditorListModel(QtCore.QAbstractListModel):
  def __init__(self, editorList: List[ParamEditor], parent: QtWidgets.QWidget=None):
    super().__init__(parent)
    self.displayFormat = '{stateName} | {editor.name}'
    self.paramStatesLst: List[str] = []
    self.editorList: List[ParamEditor] = []
    self.uniqueEditors: List[ParamEditor] = []

    self.addEditors(editorList)

  def addEditors(self, editorList: List[ParamEditor]):
    editorList = [e for e in editorList if e not in self.uniqueEditors and e.saveDir is not None]
    self.uniqueEditors.extend(editorList)
    self.layoutAboutToBeChanged.emit()
    for editor in editorList:
      for stateName in self.getParamStateFiles(editor.saveDir, editor.fileType):
        self.paramStatesLst.append(stateName)
        self.editorList.append(editor)
      editor.sigParamStateCreated.connect(lambda name, e=editor:
                                          self.addOptForEditor(e, name))
    self.layoutChanged.emit()

  def addOptForEditor(self, editor: ParamEditor, name: str):
    if self.displayFormat.format(editor=editor, stateName=name) in self.displayedData:
      return
    self.layoutAboutToBeChanged.emit()
    self.paramStatesLst.append(name)
    self.editorList.append(editor)
    self.layoutChanged.emit()

  # def updateEditorOpts(self, editor: PrjParamEditor):
  #   self.layoutAboutToBeChanged.emit()
  #   for ii in range(len(self.paramStatesLst) - 1, -1, -1):
  #     if self.editorList[ii] is editor:
  #       del self.paramStatesLst[ii]
  #       del self.editorList[ii]
  #   self.addEditors([editor])
  #   self.layoutChanged.emit()

  def data(self, index: QtCore.QModelIndex, role: int=QtCore.Qt.DisplayRole):
    row = index.row()
    paramState = self.paramStatesLst[row]
    editor = self.editorList[row]
    if role == QtCore.Qt.DisplayRole:
      return self.displayFormat.format(stateName=paramState, editor=editor)
    elif role == QtCore.Qt.EditRole:
      return paramState, editor
    else:
      return

  def stringList(self):
    return self.displayedData

  @property
  def displayedData(self):
    return [self.displayFormat.format(stateName=stng, editor=edtr)
            for stng, edtr in zip(self.paramStatesLst, self.editorList)]

  def rowCount(self, paren=QtCore.QModelIndex()) -> int:
    return len(self.paramStatesLst)

  def headerData(self, section, orientation, role=QtCore.Qt.DisplayRole):
    return 'Parameter State List'

  @staticmethod
  def getParamStateFiles(stateDir: Path, fileExt: str) -> List[str]:
    if stateDir is None: return []
    files = stateDir.glob(f'*.{fileExt}')
    return [file.stem for file in files]


def _addRmOption(param: Parameter):
  item = next(iter(param.items.keys()))

  try:
    menu: QtWidgets.QMenu = item.contextMenu
  except AttributeError: # If the menu doesn't exist yet
    menu = QtWidgets.QMenu()
    item.contextMenu = menu
  existingActs = [act.text() for act in menu.actions()]
  if 'Remove' not in existingActs:
    item.contextMenu.addAction('Remove').triggered.connect(item.requestRemove)


class QuickLoaderEditor(ParamEditor):
  def __init__(self, parent=None, editorList: List[ParamEditor]=None):
    super().__init__(parent, paramList=[], saveDir=None, fileType='loader',
                     name='Editor State Shortcuts')
    if editorList is None:
      editorList = []
    self.listModel = EditorListModel(editorList, self)
    # TODO: Get quickloader working
    self.saveAsBtn.hide()

    self.addNewParamState = uw.PopupLineEditor(self, self.listModel, clearOnComplete=False)
    self.centralLayout.insertWidget(0, self.addNewParamState)

    # self.addNewParamState.completer().activated.connect(self.addFromLineEdit)
    self.addNewParamState.returnPressed.connect(self.addFromLineEdit)

  def _loadParamState(self, stateName: Union[str, Path], stateDict: dict=None,
                     addChildren=False, removeChildren=False, applyChanges=True):
    ret = super()._loadParamState(stateName, stateDict, applyChanges=False)
    invalidGrps = []
    editorNames = [e.name for e in self.listModel.uniqueEditors]
    hasInvalidEntries = False
    for grp in self.params: # type: GroupParameter
      name = grp.name()
      # Get corresponding editor
      try:
        idx = editorNames.index(name)
        editor = self.listModel.uniqueEditors[idx]
      except ValueError:
        invalidGrps.append(grp)
        hasInvalidEntries = True
        continue
      for act in grp: # type: ShcKeySeq
        self.addActForEditor(editor, act.name(), act)
    for grp in invalidGrps:
      grp.remove()
    if hasInvalidEntries:
      errMsg = f"The following editors were not recognized:\n" \
               f"{[grp.name() for grp in invalidGrps]}\n" \
               f"Must be one of:\n" \
               f"{[e.name for e in self.listModel.uniqueEditors]}"
      warnLater(errMsg, UserWarning)
    if applyChanges:
      self.applyChanges()
    return ret

  def buildFromStartupParams(self, startupSrc: dict):
    # If quick loader is given along with other params, use the quick loader as the
    # base and apply other settings on top of it
    errSettings = []
    # Ignore case and spacing on input keys
    startupSrc = {lower_NoSpaces(kk): vv for kk, vv in startupSrc.items()}

    for editor in [self] + self.listModel.uniqueEditors: # type: ParamEditor
      paramStateInfo: Union[dict, str] = startupSrc.get(lower_NoSpaces(editor.name), None)
      try:
        if isinstance(paramStateInfo, dict):
          editor.loadParamValues(self.lastAppliedName, paramStateInfo)
        elif paramStateInfo is not None:
          editor.loadParamValues(paramStateInfo)
      except Exception as ex:
        errSettings.append(f'{editor.name}: {ex}')
    if len(errSettings) > 0:
      warnLater('The following settings could not be loaded (shown as [setting]: [exception])\n'
           + "\n\n".join(errSettings), UserWarning)
    return startupSrc

  def addDock(self, dock: Union[ParamEditor, ParamEditorDockGrouping]):
    if isinstance(dock, ParamEditorDockGrouping):
      self.listModel.addEditors(dock.editors)
    else:
      self.listModel.addEditors([dock])

  def saveCurStateAsDefault(self):
    # TODO: Get quickloader working
    return
    super().saveCurStateAsDefault()

  def saveParamValues(self, saveName: str=None, paramState: dict=None, **kwargs):
    # TODO: Get quickloader working
    # stateDict = fns.paramDictWithOpts(self.params, ['type'], [ShcKeySeq, GroupParameter])
    kwargs['paramState'] = {}
    super().saveParamValues(saveName, **kwargs)

  def applyChanges(self):
    super().applyChanges()
    for grp in self.params.childs: # type: GroupParameter
      if grp.hasChildren():
        act: ShcKeySeq = next(iter(grp))
        act.activate()

  def addFromLineEdit(self):
    completer = self.addNewParamState.completer()
    selection = completer.completionModel()
    try:
      selectionIdx = self.listModel.displayedData.index(self.addNewParamState.text())
    except ValueError:
      selectionIdx = None
    if selectionIdx is None:
      return
    qtSelectionIdx = self.listModel.index(selectionIdx)
    # selectionIdx = completer.popup().currentIndex()
    # if not selectionIdx.isValid():
    #   selectionIdx = completer.currentIndex()
    paramState, editor = qtSelectionIdx.data(QtCore.Qt.EditRole)
    self.addActForEditor(editor, paramState)
    # self.addNewParamState.clear()


  def addActForEditor(self, editor: ParamEditor, paramState: str, act: ShcKeySeq=None):
    if editor.name not in self.params.names:
      curGroup = self.params.addChild(dict(name=editor.name, type='group', removable=True))
    else:
      # Bug: If 'removable' is not specified on construction of the parameter item,
      # It is not made possible through the context menu. Fix this
      curGroup = self.params.names[editor.name]
      _addRmOption(curGroup)
      if act is None and paramState in curGroup.names:
        act = curGroup.child(paramState)

    if paramState in curGroup.names and act is not None and act.isActivateConnected:
      # Duplicate option, no reason to add
      return
    curGroup.opts['removable'] = True
    if act is None:
      act = ShcKeySeq(name=paramState, value='', removable=True, type='shortcutkeyseq')
    curGroup.addChild(act)

    act.opts['removable'] = True
    _addRmOption(act)
    act.sigActivated.connect(
      lambda _act: self._safeLoadParamValues(_act, editor, paramState))
    act.isActivateConnected = True

  def _safeLoadParamValues(self, action: ShcKeySeq, editor: ParamEditor,
                           paramState: str):
    """
    It is possible for the quick loader to refer to a param state that no longer
    exists. When this happens, failure should be graceful and the action should be
    deleted
    """
    try:
      editor.loadParamValues(paramState)
    except FileNotFoundError:
      action.remove()
      # Wait until end of process cycle to raise error
      formattedState = self.listModel.displayFormat.format(editor=editor, stateName=paramState)
      fns.raiseErrorLater(ValueError(
        f'Attempted to load {formattedState} but the setting was not found.'
      ))


from __future__ import annotations

import weakref
from dataclasses import dataclass, fields, field
from typing import Any, Optional, Collection, Union
from warnings import warn

from utilitys import PrjParam

@dataclass
class PrjParamGroup:
  """
  Hosts all child parameters and offers convenience function for iterating over them
  """

  def paramNames(self):
    """
    Outputs the column names of each parameter in the group.
    """
    return [curField.name for curField in self]

  def __iter__(self):
    # 'self' is an instance of the class, so the warning is a false positive
    # noinspection PyDataclass
    for curField in fields(self):
      yield getattr(self, curField.name)

  def __len__(self):
    return len(fields(self))

  def __str__(self):
    return f'{[f.name for f in self]}'

  def __post_init__(self):
    for param in self:
      param.group = weakref.proxy(self)

  @staticmethod
  def fieldFromParam(group: Union[Collection[PrjParam], PrjParamGroup], param: Union[str, PrjParam],
                     default: PrjParam=None):
    """
    Allows user to create a :class:`PrjParam` object from its string value (or a parameter that
    can equal one of the parameters in this list)
    """
    param = str(param).lower()
    for matchParam in group:
      if matchParam.name.lower() == param:
        return matchParam
    # If we reach here the value didn't match any CNSTomponentTypes values. Throw an error
    if default is None and hasattr(group, 'getDefault'):
      default = group.getDefault()
    baseWarnMsg = f'String representation "{param}" was not recognized.\n'
    if default is None:
      # No default specified, so we have to raise Exception
      raise ValueError(baseWarnMsg + 'No class default is specified.')
    # No exception needed, since the user specified a default type in the derived class
    warn(baseWarnMsg + f'Defaulting to {default.name}', UserWarning)
    return default

  @classmethod
  def getDefault(cls) -> Optional[PrjParam]:
    """
    Returns the default Param from the group. This can be overloaded in derived classes to yield a safe
    fallback class if the :func:`fieldFromParam` method fails.
    """
    return None

def newParam(name: str, val: Any=None, pType: str=None, helpText='', **opts):
  """
  Factory for creating new parameters within a :class:`PrjParamGroup`.

  See parameter documentation from :class:PrjParam for arguments.

  :return: Field that can be inserted within the :class:`PrjParamGroup` dataclass.
  """
  return field(default_factory=lambda: PrjParam(name, val, pType, helpText, **opts))


from typing import Callable, Dict

import numpy as np
import pyqtgraph as pg
from pyqtgraph.Qt import QtCore, QtGui
from skimage.draw import draw
from utilitys import PrjParam

from s3a.constants import PRJ_CONSTS
from s3a.structures import XYVertices, ComplexXYVertices

__all__ = ['RectROI', 'PlotDataROI', 'PolygonROI', 'PointROI', 'SHAPE_ROI_MAPPING']
from s3a.views.clickables import BoundScatterPlot

qe = QtCore.QEvent


class PlotDataROI(BoundScatterPlot):
  connected = True

  def __init__(self, endEvType=qe.MouseButtonRelease, addIntermedPts=True):
    super().__init__(pxMode=False, size=1)
    self.setBrush(pg.mkBrush(0, 0, 0, 100))
    self.setPen(pg.mkPen('w', width=0))
    self.vertices = XYVertices()

    self.endEvType = endEvType
    self.viableEvTypes = {
      qe.MouseButtonPress, qe.MouseMove, qe.MouseButtonRelease, qe.MouseButtonDblClick
    } - {endEvType}
    self.constructingRoi = False
    self.addIntermedPts = addIntermedPts
    self.firstPt = XYVertices()

  def addRoiPoints(self, pts: XYVertices):
    # noinspection PyTypeChecker
    pts = pts.astype(int)
    if self.vertices.size > 0 and np.all(pts == self.vertices[-1]):
      return
    if self.addIntermedPts or len(self.vertices) < 2:
      vertsToUse = np.vstack([self.vertices, pts])
    else:
      vertsToUse = np.vstack([self.firstPt, pts])
    refactored = self._refactorPoints(vertsToUse)
    verts = self.vertices = XYVertices(refactored)
    connectData = np.vstack([verts, verts[[0]]])
    pos = connectData.min(0, keepdims=True)
    connectData = connectData - pos + 0.5
    self.setData(*pos.T, symbol=[pg.fn.arrayToQPath(*connectData.T)])


  def setRoiPoints(self, pts: XYVertices=None):
    if pts is None:
      pts = XYVertices()
    self.setData()
    self.vertices = XYVertices()
    self.firstPt = pts
    if pts.size > 0:
      self.addRoiPoints(pts)

  def _refactorPoints(self, vertices: np.ndarray):
    return vertices

  def updateShape(self, ev: QtGui.QMouseEvent, xyEvCoords: XYVertices):
    """
    See function signature for :func:`ExtendedROI.updateShape`
    """
    success = True
    verts = None
    constructingRoi = self.constructingRoi
    # If not left click, do nothing
    if (int(ev.buttons()) & QtCore.Qt.LeftButton) == 0 \
        and ev.button() != QtCore.Qt.LeftButton:
      return self.constructingRoi, verts
    evType = ev.type()
    if evType == qe.MouseButtonPress and not constructingRoi:
      # Need to start a new shape
      self.setRoiPoints(xyEvCoords)
      constructingRoi = True
    if evType in self.viableEvTypes:
      self.addRoiPoints(xyEvCoords)
      constructingRoi = True
    elif evType == self.endEvType:
      # Done drawing the ROI, complete shape, get vertices
      verts = self.vertices
      constructingRoi = False
    else:
      success = False
      # Unable to process this event

    if success:
      ev.accept()
    self.constructingRoi = constructingRoi
    return self.constructingRoi, verts

class RectROI(PlotDataROI):

  def __init__(self, endEvType=qe.MouseButtonRelease):
    super().__init__(endEvType)
    self.addIntermedPts = False

  def _refactorPoints(self, vertices: np.ndarray):
    square = XYVertices(np.vstack([vertices.min(0), vertices.max(0)]))
    return np.vstack([
      square[0],
      [square[0,0], square[1,1]],
      [square[1,0], square[1,1]],
      [square[1,0], square[0,1]]
    ])

class PolygonROI(PlotDataROI):

  def __init__(self):
    super().__init__(qe.MouseButtonDblClick, True)
    self.viableEvTypes.remove(qe.MouseMove)
    self.viableEvTypes.remove(qe.MouseButtonRelease)

class PointROI(PlotDataROI):
  roiRadius = 1

  @classmethod
  def updateRadius(cls, radius=1):
    cls.roiRadius = radius


  def __init__(self):
    super().__init__()
    self.setSize(3)
    self.constructTypes = {qe.MouseButtonPress, qe.MouseMove}

  def updateShape(self, ev: QtGui.QMouseEvent, xyEvCoords: XYVertices):
    success = False
    verts = None
    constructingRoi = False
    if ev.type() in self.constructTypes:
      success = True
      self.setData(*xyEvCoords.T, size=self.roiRadius*2)
      verts = XYVertices(np.column_stack(draw.disk(xyEvCoords[0], self.roiRadius)))
      self.vertices = verts
      constructingRoi = True

    if success:
      ev.accept()
    return constructingRoi, verts

class EllipseROI(PlotDataROI):

  def __init__(self):
    super().__init__(endEvType=qe.MouseButtonRelease, addIntermedPts=False)

  def _refactorPoints(self, vertices: np.ndarray):
    pts = vertices[:, ::-1]
    ma = pts.max(0).astype(int)
    mi = pts.min(0).astype(int)
    center = (ma - mi)//2 + mi
    # Ideally, the perimeter could be directly returned. However, the ordering is not
    # compatible with direct column stacking (results in criss-crossing)
    filled = draw.ellipse_perimeter(*center, ma[0]-mi[0], ma[1]-mi[1])
    if filled[0].size <= 1:
      return vertices
    mask = np.zeros((filled[0].max()+1, filled[1].max()+1), bool)
    mask[filled[0], filled[1]] = True
    return ComplexXYVertices.fromBwMask(mask, externOnly=True).stack()

SHAPE_ROI_MAPPING: Dict[PrjParam, Callable[[], PlotDataROI]] = {
  PRJ_CONSTS.DRAW_SHAPE_RECT: RectROI,
  PRJ_CONSTS.DRAW_SHAPE_FREE: PlotDataROI,
  PRJ_CONSTS.DRAW_SHAPE_POLY: PolygonROI,
  PRJ_CONSTS.DRAW_SHAPE_ELLIPSE: EllipseROI,
  PRJ_CONSTS.DRAW_SHAPE_POINT: PointROI,
}



from __future__ import annotations

import inspect
import os
import sys

import pyqtgraph as pg
from pyqtgraph.Qt import QtWidgets

__all__ = ['appInst', 'PRJ_SINGLETON', 'S3A', 'REQD_TBL_FIELDS',
           'ComplexXYVertices', 'XYVertices', 'PRJ_CONSTS', 'ComponentIO',
           '__version__']

pg.setConfigOptions(imageAxisOrder='row-major')

from .__version__ import __version__

# Makes sure that when the folder is run as a module, the app exists in the outermost
# scope of the application
customPlatform = os.environ.get('S3A_PLATFORM')
appInst = QtWidgets.QApplication.instance()
if appInst is None:
  args = list(sys.argv)
  if customPlatform is not None:
    args += ['-platform', customPlatform]
  appInst = QtWidgets.QApplication(args)
# Now that the app was created with sys args, populate pg instance
pg.mkQApp()
# Allow selectable text in message boxes
appInst.setStyleSheet("QMessageBox { messagebox-text-interaction-flags: 5; }")

from . import graphicsutils as gutils

# appInst.installEventFilter(
#   gutils.QAwesomeTooltipEventFilter(appInst))

# Import here to resolve resolution order
import s3a.constants
import s3a.structures

from s3a.parameditors import PRJ_SINGLETON
from ._io import ComponentIO, defaultIo
from s3a.processing.algorithms import TopLevelImageProcessors, TOP_GLOBAL_PROCESSOR_FUNCS
from s3a.structures import XYVertices, ComplexXYVertices
from s3a.constants import REQD_TBL_FIELDS, PRJ_CONSTS

# -----
# DEFAULT PLUGINS
# -----
for name, func in inspect.getmembers(TopLevelImageProcessors, inspect.isfunction):
  PRJ_SINGLETON.imgProcClctn.addProcessCtor(func)

for ctor in TOP_GLOBAL_PROCESSOR_FUNCS:
  PRJ_SINGLETON.multiPredClctn.addProcessCtor(ctor)

from .views.s3agui import S3A

from s3a.plugins import ALL_PLUGINS
from s3a.plugins.misc import RandomToolsPlugin
for plugin in ALL_PLUGINS():
  PRJ_SINGLETON.addPlugin(plugin)

miscPlugin: RandomToolsPlugin = PRJ_SINGLETON.clsToPluginMapping[RandomToolsPlugin]


from __future__ import annotations

import traceback
from copy import deepcopy
from typing import List

import numpy as np
from utilitys import NestedProcWrapper
from utilitys.fns import warnLater

from s3a.generalutils import augmentException
from s3a.processing.algorithms import crop_to_local_area, apply_process_result, \
  basicOpsCombo, return_to_full_size, format_vertices
from s3a.structures import ComplexXYVertices, XYVertices
from .processing import *

__all__ = ['ImgProcWrapper', 'NestedProcWrapper']

def _prependFuncs():
  formatStage = ImageProcess.fromFunction(format_vertices)
  formatStage.allowDisable = False
  cropStage = ImageProcess.fromFunction(crop_to_local_area)
  return [formatStage, cropStage]

def _appendFuncs():
  applyStage = ImageProcess.fromFunction(apply_process_result)
  applyStage.allowDisable = False
  resizeStage = ImageProcess.fromFunction(return_to_full_size)
  resizeStage.allowDisable = False

  return [applyStage, basicOpsCombo(), resizeStage]

class ImgProcWrapper(NestedProcWrapper):
  preProcStages = _prependFuncs()
  postProcStages = _appendFuncs()

  def __init__(self, processor: ImageProcess, *,
               excludedStages: List[List[str]]=None, disabledStages: List[List[str]]=None,
               **kwargs):
    # Each processor is encapsulated in processes that crop the image to the region of
    # interest specified by the user, and re-expand the area after processing
    self.output = np.zeros((0,0), bool)

    preStages = [*map(deepcopy, self.preProcStages)]
    finalStages = [*map(deepcopy, self.postProcStages)]
    processor.stages = preStages + processor.stages + finalStages

    if disabledStages is None:
      disabledStages = []
    if hasattr(processor, 'disabledStages'):
      disabledStages.extend(processor.disabledStages)

    if excludedStages is None:
      excludedStages = []
    if hasattr(processor, 'excludedStages'):
      excludedStages.extend(processor.excludedStages)

    for namePath in disabledStages: # type: List[str]
      proc = self.getNestedName(processor, namePath)
      proc.disabled = True
    for namePath in excludedStages: # type: List[str]
      parentProc = self.getNestedName(processor, namePath[:-1])
      parentProc.stages.remove(self.getNestedName(parentProc, [namePath[-1]]))
    super().__init__(processor, **kwargs)

  def run(self, **kwargs):
    newIo = self._ioDictFromRunKwargs(kwargs)

    try:
      result = self.processor.run(newIo)
    except Exception as ex:
      augmentException(ex, 'Exception during processor run:\n'
                       + ''.join(traceback.format_stack(limit=5)))
      result = ProcessIO(image=kwargs['prevCompMask'])
      warnLater(str(ex), UserWarning)

    outImg = result['image'].astype(bool)
    if outImg.ndim > 2:
      outImg = np.bitwise_or.reduce(outImg, 2)
    self.output = outImg
    return outImg

  def resultAsVerts(self, localEstimate=True):
    initialList = ComplexXYVertices.fromBwMask(self.output)
    if len(initialList) == 0:
      return initialList
    if not localEstimate:
      return [ComplexXYVertices([lst]) for lst in initialList]
    # else, all vertices belong to the same component
    else:
      return [initialList]

  @staticmethod
  def _ioDictFromRunKwargs(runKwargs):
    image = runKwargs.get('image', None)
    if image is None:
      raise RuntimeError('Cannot run processor without an image')

    runKwargs.setdefault('firstRun', True)
    for name in 'fgVerts', 'bgVerts':
      runKwargs.setdefault(name, XYVertices())

    if runKwargs.get('prevCompMask', None) is None:
      noPrevMask = True
      runKwargs['prevCompMask'] = np.zeros(image.shape[:2], bool)
    else:
      noPrevMask = False

    return ProcessIO(**runKwargs, noPrevMask=noPrevMask)


from __future__ import annotations

from typing import Tuple, Sequence, Optional, List

import numpy as np
import pandas as pd
import pyqtgraph as pg
from matplotlib import cm
from matplotlib.pyplot import colormaps
from pandas import DataFrame as df
from pyqtgraph import arrayToQPath
from pyqtgraph.Qt import QtGui, QtCore
from utilitys import PrjParam, RunOpts, EditorPropsMixin, fns

from s3a import PRJ_SINGLETON
from s3a.constants import REQD_TBL_FIELDS as RTF, PRJ_CONSTS, PRJ_ENUMS
from s3a.generalutils import stackedVertsPlusConnections
from s3a.structures import GrayImg, OneDArr, BlackWhiteImg
from s3a.structures import XYVertices, ComplexXYVertices
from . import imageareas
from .clickables import BoundScatterPlot

__all__ = ['MultiRegionPlot', 'VertexDefinedImg', 'RegionCopierPlot']

from .._io import defaultIo

Signal = QtCore.Signal

def makeMultiRegionDf(numRows=1, idList: Sequence[int]=None, selected:Sequence[bool]=None,
                      focused: Sequence[bool]=None,
                      vertices: Sequence[ComplexXYVertices]=None, lblField: PrjParam=None):
  """
  Helper for creating new dataframe holding information determining color data.
  `selected` and `focused` must be boolean arrays indicating whether or not each component
  is selected or focused, respectively.
  If `lblField` is given, it is used as a value to color the components
  """
  outDict = {}
  if selected is None:
    selected = np.zeros(numRows, bool)
  outDict[PRJ_ENUMS.FIELD_SELECTED] = selected
  if focused is None:
    focused = np.zeros(numRows, bool)
  outDict[PRJ_ENUMS.FIELD_FOCUSED] = focused
  if lblField is not None:
    labels_tmp = np.tile(lblField.value, numRows)
    labels = lblField.toNumeric(labels_tmp, rescale=True)
  else:
    labels = np.zeros(numRows)
  outDict[PRJ_ENUMS.FIELD_LABEL] = labels
  if vertices is None:
    vertices = [RTF.VERTICES.value for _ in range(numRows)]
  outDict[RTF.VERTICES] = vertices
  outDf = pd.DataFrame(outDict)
  if idList is not None:
    outDf = outDf.set_index(idList, drop=True)
  return outDf

def _makeTxtSymbol(txt: str, fontSize: int):
  outSymbol = QtGui.QPainterPath()
  txtLabel = QtGui.QFont("Sans Serif", fontSize)
  txtLabel.setStyleStrategy(QtGui.QFont.PreferBitmap | QtGui.QFont.PreferQuality)
  outSymbol.addText(0, 0, txtLabel, txt)
  br = outSymbol.boundingRect()
  scale = min(1. / br.width(), 1. / br.height())
  tr = QtGui.QTransform()
  tr.scale(scale, scale)
  tr.translate(-br.x() - br.width()/2., -br.y() - br.height()/2.)
  outSymbol = tr.map(outSymbol)
  return outSymbol

def _makeBoundSymbol(verts: XYVertices):
  verts = verts - verts.min(0, keepdims=True)
  path = arrayToQPath(*verts.T, connect='finite')
  return path

class MultiRegionPlot(EditorPropsMixin, BoundScatterPlot):
  __groupingName__ = PRJ_CONSTS.CLS_MULT_REG_PLT.name

  @classmethod
  def __initEditorParams__(cls):
    super().__initEditorParams__()

  def __init__(self, parent=None):
    super().__init__(size=1, pxMode=False)
    # Wrapping in atomic process means when users make changes to properties, these are maintained when calling the
    # function internally with no parameters
    with PRJ_SINGLETON.colorScheme.setBaseRegisterPath(self.__groupingName__):
      self.updateColors = PRJ_SINGLETON.colorScheme.registerFunc(
        self.updateColors, runOpts=RunOpts.ON_CHANGED, nest=False, ignoreKeys=['hideFocused'])
    self.setParent(parent)
    self.setZValue(50)
    self.regionData = makeMultiRegionDf(0)
    self.cmap = np.array([])
    self.updateColors()

    # 'pointsAt' is an expensive operation if many points are in the scatterplot. Since
    # this will be called anyway when a selection box is made in the main image, disable
    # mouse click listener to avoid doing all that work for nothing.
    # self.centroidPlts.mouseClickEvent = lambda ev: None
    self.mouseClickEvent = lambda ev: None
    # Also disable sigClicked. This way, users who try connecting to this signal won't get
    # code that runs but never triggers
    # self.centroidPlts.sigClicked = None
    self.sigPointsClicked = None

  def resetRegionList(self, newRegionDf: Optional[df]=None,
                      lblField:PrjParam=RTF.INST_ID):
    idList = None
    if (newRegionDf is not None
        and lblField in newRegionDf.columns):
      newRegionDf = newRegionDf.copy()
      newRegionDf[PRJ_ENUMS.FIELD_LABEL] = lblField.toNumeric(newRegionDf[lblField], rescale=True)
    numRows = len(newRegionDf)
    if newRegionDf is not None:
      idList = newRegionDf.index
    self.regionData = makeMultiRegionDf(numRows, idList=idList, lblField=lblField)
    if newRegionDf is not None:
      self.regionData.update(newRegionDf)
    self.updatePlot()

  def selectById(self, selectedIds: OneDArr):
    """
    Marks 'selectedIds' as currently selected by changing their scheme to user-specified
    selection values.
    """
    self.updateSelected_focused(selectedIds=selectedIds)

  def focusById(self, focusedIds: OneDArr):
    """
    Colors 'focusedIds' to indicate they are present in a focused views.
    """
    self.updateSelected_focused(focusedIds=focusedIds)

  def updateSelected_focused(self, selectedIds:np.ndarray=None,
                             focusedIds: np.ndarray=None):
    """
    :param selectedIds: All currently selected Ids
    :param focusedIds: All currently focused Ids
    """
    if len(self.regionData) == 0:
      return
    for col, idList in zip(self.regionData.columns, [selectedIds, focusedIds]):
      if idList is None: continue
      self.regionData[col] = False
      idList = np.intersect1d(self.regionData.index, idList)
      self.regionData.loc[idList, col] = True
    self.updateColors()

  def updatePlot(self):
    # -----------
    # Update data
    # -----------
    boundLocs = []
    boundSymbs = []
    if self.regionData.empty:
      self.setData(x=[], y=[], data=[])
      return

    for region, _id in zip(self.regionData[RTF.VERTICES],
                           self.regionData.index):
      concatRegion, isfinite = stackedVertsPlusConnections(region)
      boundLoc = np.nanmin(concatRegion, 0, keepdims=True)
      boundSymbol = pg.arrayToQPath(*(concatRegion-boundLoc+0.5).T, connect=isfinite)

      boundLocs.append(boundLoc)
      boundSymbs.append(boundSymbol)

    plotRegions = np.vstack(boundLocs)
    self.setData(*plotRegions.T, symbol=boundSymbs,
                          data=self.regionData.index)
    self.updateColors()

  def toGrayImg(self, imShape: Sequence[int]=None):
    uint16Max = 2**16-2 # Subtract 1 extra so there's room for offset
    labels = (self.regionData[PRJ_ENUMS.FIELD_LABEL]*uint16Max).astype('uint16')
    labelDf = pd.DataFrame()
    labelDf[RTF.VERTICES] = self.regionData[RTF.VERTICES]
    # Override id column to avoid an extra parameter
    labelDf[RTF.INST_ID] = labels
    return defaultIo.exportLblPng(labelDf, imShape=imShape, allowOffset=True)

  @fns.dynamicDocstring(cmapVals=colormaps() + ['None'])
  def updateColors(self, penWidth=0, penColor='w', selectedFill='00f', focusedFill='f00', labelColormap='viridis',
                   fillAlpha=0.7):
    """
    Assigns colors from the specified colormap to each unique class
    :param penWidth: Width of the pen in pixels
    :param penColor:
      helpText: Color of the border of each non-selected boundary
      pType: color
    :param selectedFill:
      helpText: Fill color for components selected in the component table
      pType: color
    :param focusedFill:
      helpText: Fill color for the component currently in the focused image
      pType: color
    :param labelColormap:
      helpText: "Colormap to use for fill colors by component label. If `None` is selected,
        the fill will be transparent."
      pType: popuplineeditor
      limits: {cmapVals}
    :param fillAlpha:
      helpText: Transparency of fill color (0 is totally transparent, 1 is totally opaque)
      limits: [0,1]
      step: 0.1
    :param hideFocused: Many plugins alter the visual behavior of focused regions, so it
      is helpful to have a flag which will cause focused regions to be hidden.
    """
    if len(self.regionData) == 0 or len(self.data) == 0:
      return
    def colorsWithAlpha(_numericLbls: np.ndarray):
      useAlpha = fillAlpha
      if labelColormap == 'None':
        cmap = lambda _classes: np.zeros((len(_classes), 4))
        useAlpha = 0.0
      else:
        cmap = cm.get_cmap(labelColormap)
      colors = cmap(_numericLbls)
      colors[:,-1] = useAlpha
      return colors
    selectedFill = pg.Color(selectedFill)
    focusedFill = pg.Color(focusedFill)

    focusedFill = np.array(pg.Color(focusedFill).getRgbF())
    focusedFill[-1] = fillAlpha
    selectedFill = np.array(pg.Color(selectedFill).getRgbF())
    selectedFill[-1] = fillAlpha
    # combinedFill = (focusedFill + selectedFill)/2
    lbls = self.regionData[PRJ_ENUMS.FIELD_LABEL].to_numpy()
    fillColors = colorsWithAlpha(lbls)
    selected = self.regionData[PRJ_ENUMS.FIELD_SELECTED].to_numpy()
    focused = self.regionData[PRJ_ENUMS.FIELD_FOCUSED].to_numpy()
    fillColors[selected] = selectedFill
    fillColors[focused] = focusedFill
    # fillColors[focused & selected] = combinedFill
    penColors = np.tile(pg.mkPen(color=penColor, width=penWidth), len(lbls))
    self.setPen(penColors)
    self.setBrush([pg.Color(f*255) for f in fillColors])
    self.invalidate()

  def drop(self, ids):
    self.regionData.drop(index=ids, inplace=True)

  def dataBounds(self, ax, frac=1.0, orthoRange=None):
    allVerts = ComplexXYVertices()
    for v in self.regionData[RTF.VERTICES]:
      allVerts.extend(v)
    allVerts = allVerts.stack()
    if len(allVerts) == 0:
      return [None, None]
    bounds = np.r_[allVerts.min(0, keepdims=True)-0.5, allVerts.max(0, keepdims=True)+0.5]
    return list(bounds[:,ax])


class VertexDefinedImg(EditorPropsMixin, pg.ImageItem):
  sigRegionReverted = Signal(object) # new GrayImg

  __groupingName__ = 'Focused Image Graphics'

  @classmethod
  def __initEditorParams__(cls):
    cls.fillClr, cls.vertClr = PRJ_SINGLETON.colorScheme.registerProps(
      [PRJ_CONSTS.SCHEME_REG_FILL_COLOR, PRJ_CONSTS.SCHEME_REG_VERT_COLOR])

  def __init__(self):
    super().__init__()
    self.verts = ComplexXYVertices()
    PRJ_SINGLETON.colorScheme.sigChangesApplied.connect(lambda: self.setImage(
      lut=self.getLUTFromScheme()))

  def embedMaskInImg(self, toEmbedShape: Tuple[int, int]):
    outImg = np.zeros(toEmbedShape, dtype=bool)
    selfShp = self.image.shape
    outImg[0:selfShp[0], 0:selfShp[1]] = self.image
    return outImg

  @PRJ_SINGLETON.actionStack.undoable('Modify Focused Region')
  def updateFromVertices(self, newVerts: ComplexXYVertices, srcImg: GrayImg=None):
    oldImg = self.image
    oldVerts = self.verts

    self.verts = newVerts.copy()
    if len(newVerts) == 0:
      regionData = np.zeros((1, 1), dtype=bool)
    else:
      if srcImg is None:
        stackedVerts = newVerts.stack()
        regionData = newVerts.toMask(asBool=False)
        # Make vertices full brightness
        regionData[stackedVerts.rows, stackedVerts.cols] = 2
      else:
        regionData = srcImg.copy()

    self.setImage(regionData, levels=[0, 2], lut=self.getLUTFromScheme())
    yield
    self.updateFromVertices(oldVerts, oldImg)

  def updateFromMask(self, newMask: BlackWhiteImg):
    # It is expensive to color the vertices, so only find contours if specified by the user
    oldImg = self.image
    oldVerts = self.verts

    newMask = newMask.astype('uint8')
    if np.array_equal(oldImg>0, newMask):
      # Nothing to do
      return
    verts = ComplexXYVertices.fromBwMask(newMask)
    stackedVerts = verts.stack()
    newMask[stackedVerts.rows, stackedVerts.cols] = 2
    self.updateFromVertices(verts, srcImg=newMask)
    return

  def getLUTFromScheme(self):
    lut = [(0, 0, 0, 0)]
    for clr in self.fillClr, self.vertClr:
      lut.append(clr.getRgb())
    return np.array(lut, dtype='uint8')

class RegionCopierPlot(pg.PlotCurveItem):
  sigCopyStarted = QtCore.Signal()
  sigCopyStopped = QtCore.Signal()

  def __init__(self, mainImg: imageareas.MainImage=None, parent=None):
    super().__init__(parent)
    self.active = False
    self.inCopyMode = True
    self.baseData = XYVertices()
    self.regionIds = np.ndarray([])
    self.dataMin = XYVertices()
    self.offset = XYVertices([[0,0]])

    self.setShadowPen(color='k', width=2*self.opts['pen'].width())
    """
    Instead of a customizeable color palette for the copy shape, it is easier to
    have a black outline and white inline color for the shape plot which ensures
    all vertices are visible on any background. However, it is not easy to create
    a multicolored pen in pyqt -- the much simpler solution is to simply create
    a shadow pen, where one has a boundary twice as thick as the other.
    """

    self._connectivity = np.ndarray([], bool)
    mainImg.sigMouseMoved.connect(self.mainMouseMoved)

  def mainMouseMoved(self, xyPos: np.ndarray):
    if not self.active: return
    newData = self.baseData + xyPos
    self.setData(newData[:,0], newData[:,1], connect=self._connectivity)
    self.offset = xyPos - self.dataMin

  def resetBaseData(self, baseData: List[ComplexXYVertices], regionIds: OneDArr):
    allData = ComplexXYVertices()
    allConnctivity = []
    for verts in baseData: # each list element represents one component
      plotData, connectivity = stackedVertsPlusConnections(verts)
      allData.append(plotData)
      allConnctivity.append(connectivity)
      if len(connectivity) > 0:
        connectivity[-1] = False
    plotData = allData.stack()
    connectivity = np.concatenate(allConnctivity)

    try:
      allMin = plotData.min(0)
      closestPtIdx  = np.argmin(np.sum(np.abs(plotData - allMin), 1))
      # Guarantees that the mouse will be on the boundary closest to the top left
      self.dataMin = plotData[closestPtIdx]
      # connectivity[addtnlFalseConnectivityIdxs] = False
    except ValueError:
      # When no elements are in the array
      self.dataMin = XYVertices([[0,0]])
    baseData: XYVertices = plotData - self.dataMin
    self.baseData = baseData
    self._connectivity = connectivity
    self.setData(plotData[:,0], plotData[:,1], connect=connectivity)

    self.regionIds = regionIds

  def erase(self):
    self.resetBaseData([ComplexXYVertices()], np.array([]))
    self.active = False


import sys
from typing import Optional

import fire
from pyqtgraph.Qt import QtCore

from . import appInst, PRJ_SINGLETON
from .views.s3agui import S3A
from utilitys.fns import makeExceptionsShowDialogs


def main(loadLastState=True, help=False, **profileArgs):
  """
  Calling code for the S3A application.

  :param loadLastState: When the app is closed, all settings are saved. If this is *True*,
    these settings are restored on startup. If *False*, they aren't.

  :param help: Show help documentation

  :key image: Optional initial image to be annotated
  :key annotations: Optional initial annotation file loaded.
  :key `param editor name`: Name of the parameter editor within S3A with a loadable state.
    This can be e.g. `colorscheme`, `shortcuts`, etc.
  """
  # Handle here for faster bootup
  win = S3A(guiMode=not help, loadLastState=loadLastState, **profileArgs)
  if help:
    return win
  makeExceptionsShowDialogs(win)
  QtCore.QTimer.singleShot(0, win.showMaximized)
  sys.exit(appInst.exec_())

if __name__ == '__main__':
    fire.Fire(main)


from __future__ import annotations

from functools import wraps
from pathlib import Path
from typing import Dict, Union, List

import numpy as np
from pyqtgraph.Qt import QtWidgets, QtCore, QtGui
import pyqtgraph as pg

from utilitys import RunOpts
from utilitys.widgets import ImageViewer, EasyWidget

Signal = QtCore.Signal
QCursor = QtGui.QCursor

def disableAppDuringFunc(func):
  @wraps(func)
  def disableApp(*args, **kwargs):
    # Captures 'self' instance
    mainWin = args[0]
    try:
      mainWin.setEnabled(False)
      return func(*args, **kwargs)
    finally:
      mainWin.setEnabled(True)
  return disableApp

def create_addMenuAct(mainWin: QtWidgets.QWidget, parentMenu: QtWidgets.QMenu, title: str, asMenu=False) \
    -> Union[QtWidgets.QMenu, QtWidgets.QAction]:
  menu = None
  if asMenu:
    menu = QtWidgets.QMenu(title, mainWin)
    act = menu.menuAction()
  else:
    act = QtWidgets.QAction(title)
  parentMenu.addAction(act)
  if asMenu:
    return menu
  else:
    return act


def autosaveOptsDialog(parent):
  dlg = QtWidgets.QDialog(parent)
  layout = QtWidgets.QGridLayout(dlg)
  dlg.setLayout(layout)

  fileBtn = QtWidgets.QPushButton('Select Output Folder')
  folderNameDisplay = QtWidgets.QLabel('', dlg)
  def retrieveFolderName():
    folderDlg = QtWidgets.QFileDialog(dlg)
    folderDlg.setModal(True)
    folderName = folderDlg.getExistingDirectory(dlg, 'Select Output Folder')

    dlg.folderName = folderName
    folderNameDisplay.setText(folderName)
  fileBtn.clicked.connect(retrieveFolderName)

  layout.addWidget(folderNameDisplay, 0, 0)
  layout.addWidget(fileBtn, 0, 1)

  saveLbl = QtWidgets.QLabel('Save Name:', dlg)
  baseFileNameEdit = QtWidgets.QLineEdit(dlg)
  dlg.baseFileNameEdit = baseFileNameEdit

  layout.addWidget(saveLbl, 1, 0)
  layout.addWidget(baseFileNameEdit, 1, 1)

  intervalLbl = QtWidgets.QLabel('Save Interval (mins):', dlg)
  intervalEdit = QtWidgets.QSpinBox(dlg)
  intervalEdit.setMinimum(1)
  dlg.intervalEdit = intervalEdit

  layout.addWidget(intervalLbl, 2, 0)
  layout.addWidget(intervalEdit, 2, 1)

  saveDescr = QtWidgets.QLabel('Every <em>interval</em> minutes, a new autosave is'
                               ' created from the component list. Its name is'
                               ' [Parent Folder]/[base name]_[counter].csv, where'
                               ' counter is the current save file number.', dlg)
  saveDescr.setWordWrap(True)
  layout.addWidget(saveDescr, 3, 0, 1, 2)


  okBtn = QtWidgets.QPushButton('Ok')
  cancelBtn = QtWidgets.QPushButton('Cancel')
  cancelBtn.clicked.connect(dlg.reject)
  okBtn.clicked.connect(dlg.accept)
  dlg.okBtn = okBtn

  layout.addWidget(okBtn, 4, 0)
  layout.addWidget(cancelBtn, 4, 1)
  return dlg

class ThumbnailViewer(QtWidgets.QListWidget):
  sigDeleteRequested = QtCore.Signal(object)
  """List[Selected image paths]"""
  sigImageSelected = QtCore.Signal(object)
  """Full path of selected image"""

  def __init__(self, parent=None):
    super().__init__(parent)
    self.nameToFullPathMapping: Dict[str, Path] = {}
    self.setViewMode(self.IconMode)
    self.setIconSize(QtCore.QSize(200,200))
    self.setResizeMode(self.Adjust)
    self.itemActivated.connect(lambda item: self.sigImageSelected.emit(self.nameToFullPathMapping[item.text()]))

    # def findDelImgs():
    #   selection = self.selectedImages
    #   self.sigDeleteRequested.emit(selection)
    # TODO: Incorporate deletion. Just using "delete" button isn't great since
    #   when shortcuts conflict it is ambiguous as to whether an image will
    #   actually be deleted
    # self.delShc = QtWidgets.QShortcut(QtCore.Qt.Key_Delete, self, findDelImgs)

  def addThumbnail(self, fullName: Path, force=False):
    icon = QtGui.QIcon(str(fullName))
    if fullName.name in self.nameToFullPathMapping:
      if force:
        self.removeThumbnail(fullName.name)
      else:
        raise IOError('Name already exists in image list')
    newItem = QtWidgets.QListWidgetItem(fullName.name)
    newItem.setIcon(icon)
    self.addItem(newItem)
    self.nameToFullPathMapping[fullName.name] = fullName

  @property
  def selectedImages(self):
    return [self.nameToFullPathMapping[idx.data()] for idx in self.selectedIndexes()]

  def removeThumbnail(self, name: str):
    items = self.findItems(name, QtCore.Qt.MatchExactly)
    if len(items) == 0:
      return
    item = items[0]
    self.takeItem(self.row(item))
    del self.nameToFullPathMapping[name]

  def clear(self):
    super().clear()
    self.nameToFullPathMapping.clear()

# Taken directly from https://stackoverflow.com/questions/60663793/drop-one-or-more-files-into-listwidget-or-lineedit
class DropList(QtWidgets.QListWidget):
  def __init__(self, parent=None):
    super(DropList, self).__init__(parent)
    self.setAcceptDrops(True)
    self.setSelectionMode(self.ExtendedSelection)
    self.delShc = QtWidgets.QShortcut(QtCore.Qt.Key_Delete, self, self.deleteSelected)

  def deleteSelected(self):
    selectedIdxs = self.selectionModel().selectedIndexes()
    selectedRows = reversed(sorted([i.row() for i in selectedIdxs]))
    for row in selectedRows:
      self.takeItem(row)

  def dragEnterEvent(self, event):
    if event.mimeData().hasUrls():
      event.acceptProposedAction()
    else:
      event.ignore()

  def dragMoveEvent(self, event):
    if event.mimeData().hasUrls():
      event.acceptProposedAction()
    else:
      event.ignore()

  def dropEvent(self, event):
    md = event.mimeData()
    if md.hasUrls():
      for url in md.urls():
        self.addItem(url.toLocalFile())
      event.acceptProposedAction()

  @property
  def files(self):
    model = self.model()
    return [model.index(ii, 0).data() for ii in range(model.rowCount())]

class RegionHistoryViewer(QtWidgets.QMainWindow):
  sigDiffsChanged = QtCore.Signal()

  def __init__(self, parent=None):
    super().__init__(parent)
    self.diffs: List[np.ndarray] = []
    self.histTimer = QtCore.QTimer(self)

    self.diffImg = pg.ImageItem()
    dp = self.displayPlt = ImageViewer()
    self.displayPlt.addItem(self.diffImg)
    self.diffImg.setOpacity(0.5)

    _, param = dp.toolsEditor.registerFunc(self.updateImg, runOpts=RunOpts.ON_CHANGED,
                                       returnParam=True)
    self.slider = param.child('curSlice')
    self.sigDiffsChanged.connect(lambda: self.slider.setLimits([0, len(self.diffs)-1]))
    self.histTimer.timeout.connect(self.incrSlicer)

    dp.toolsEditor.registerFunc(self.autoPlay)
    dp.toolsEditor.registerFunc(lambda: self.histTimer.stop(), name='Stop Autoplay')
    dp.toolsEditor.registerFunc(self.discardLeftEntries, name='Discard Entries Left of Slider')

    EasyWidget.buildMainWin([dp], layout='H', win=self)
    self.addDockWidget(QtCore.Qt.RightDockWidgetArea, dp.toolsEditor)


  def setDiffs(self, diffs: List[np.ndarray]):
    self.diffs = diffs
    self.sigDiffsChanged.emit()

  def autoPlay(self, timestep=500):
    """
    :param timestep:
      title: Timestep (ms)
    """
    self.histTimer.start(timestep)

  def incrSlicer(self):
    if self.slider.value() < len(self.diffs)-1:
      self.slider.setValue(self.slider.value()+1)
    else:
      self.histTimer.stop()

  def show(self):
    super().show()
    self.displayPlt.toolsEditor.show()

  def updateImg(self, curSlice=0):
    """
    :param curSlice:
      pType: slider
      limits: [0, 10]
    :return:
    """
    self.diffImg.setImage(self.diffs[curSlice])

  def discardLeftEntries(self):
    self.setDiffs(self.diffs[self.slider.value():])


from __future__ import annotations

from ast import literal_eval
from typing import Union, List, Sequence
from warnings import warn

import cv2 as cv
import numpy as np

from .typeoverloads import NChanImg, BlackWhiteImg

class XYVertices(np.ndarray):
  connected = True

  def __new__(cls, inputArr: Union[list, np.ndarray, tuple]=None, connected=True, dtype=int,
              **kwargs):
    # Default to integer type if not specified, since this is how pixel coordinates will be represented anyway
    # See numpy docs on subclassing ndarray
    if inputArr is None:
      inputArr = np.zeros((0,2))
    arr = np.asarray(inputArr, dtype=dtype, **kwargs).view(cls)
    arr.connected = connected
    return arr

  def __array_finalize__(self, obj):
    shape = self.shape
    shapeLen = len(shape)
    # indicates point, so the one dimension must have only 2 elements
    if 1 < shapeLen < 2 and shape[0] != 2:
      raise ValueError(f'A one-dimensional vertex array must be shape (2,).'
                                f' Receieved array of shape {shape}')
    elif shapeLen > 2 or shapeLen > 1 and shape[1] != 2:
      raise ValueError(f'Vertex list must be Nx2. Received shape {shape}.')
    if obj is None: return
    self.connected = getattr(obj, 'connected', True)

  @property
  def empty(self):
      return len(self) == 0

  def serialize(self):
    return str(self.tolist())

  @classmethod
  def deserialize(cls, strVal: str):
    out = cls(literal_eval(strVal))
    if out.size == 0:
      # Make sure size is 0x2
      return cls()
    return out

  def asPoint(self):
    if self.size == 2:
      return self.reshape(-1)
    # Reaching here means the user requested vertices as point when
    # more than one point is in the list
    raise ValueError(f'asPoint() can only be called when one vertex is in'
                              f' the vertex list. Currently has shape {self.shape}')

  def asRowCol(self):
    return np.fliplr(self)

  @property
  def x(self):
    # Copy to array first so dimensionality checks are no longer required
    return np.array(self).reshape(-1,2)[:,[0]]
  @x.setter
  def x(self, newX): self.reshape(-1,2)[:,0] = newX

  @property
  def y(self):
    return np.array(self).reshape(-1,2)[:,[1]]
  @y.setter
  def y(self, newY): self.reshape(-1,2)[:,1] = newY

  @property
  def rows(self): return self.y
  @rows.setter
  def rows(self, newRows): self.y = newRows

  @property
  def cols(self):return self.x
  @cols.setter
  def cols(self, newCols):self.x = newCols


class ComplexXYVertices(list):
  """
  Allows holes in the component shape. Subclassing ndarray instead of list allows primitive algebraic ops on the list
  contents (e.g. subtracting/adding offset). Since normal usage doesn't typically require a mutable structure, the
  loss is minimal.
  """
  hierarchy = np.ones((0,4), dtype=int)
  """See cv.findContours for hierarchy explanation. Used in cv.RETR_CCOMP mode."""

  def __init__(self, inputArr: Union[List[XYVertices], np.ndarray]=None,
               hierarchy:np.ndarray=None,
               coerceListElements=False):

    if inputArr is None:
      inputArr = []
    numInpts = len(inputArr)
    if coerceListElements:
      inputArr = [XYVertices(el) for el in inputArr]
    if hierarchy is None:
      hierarchy = np.ones((numInpts, 4), dtype=int)*-1
    super().__init__(inputArr)
    # No hierarchy required unless list is longer than length 1
    self.hierarchy = hierarchy

  def append(self, verts:XYVertices=None) -> None:
    if verts is not None:
      super().append(verts)

  def isEmpty(self):
    return len(self.stack()) == 0

  @property
  def x(self):
    return [lst.x for lst in self]
  @x.setter
  def x(self, newX):
    for lst, newLstX in zip(self, newX):
      lst.x = newLstX

  @property
  def y(self):
    return [lst.y for lst in self]

  @y.setter
  def y(self, newY):
    for lst, newLstY in zip(self, newY):
      lst.y = newLstY

  def asPoint(self):
    if len(self) == 1:
      return self[0].asPoint()
    else:
      raise ValueError(f'Can only treat ComplexXYVertices with one inner list as a point.'
                                f' Current list has {len(self)} elements.')

  def stack(self, newDtype=int) -> XYVertices:
    if len(self) == 0:
      # Check required for np vstack since it won't work with a 0-element array
      return XYVertices()
    else:
      return XYVertices(np.vstack(self), dtype=newDtype)

  @classmethod
  def stackedMax(cls, complexVertList: list):
    """
    Returns the max along dimension 0 for a list of complex vertices
    """
    return np.vstack([v.stack() for v in complexVertList]).max(0)

  def filledVerts(self) -> ComplexXYVertices:
    """
    Retrieves all vertex lists corresponding to filled regions in the complex shape
    """
    idxs = np.nonzero(self.hierarchy[:,3] == -1)[0]
    return ComplexXYVertices([self[ii] for ii in idxs])

  def holeVerts(self) -> ComplexXYVertices:
    """
    Retrieves all vertex lists corresponding to holes in the complex shape
    """
    idxs = np.nonzero(self.hierarchy[:,3] != -1)[0]
    return ComplexXYVertices([self[ii] for ii in idxs])

  def toMask(self, maskShape: Union[Sequence, NChanImg]=None,
             fillColor: Union[int, float, np.ndarray]=None,
             asBool=True, checkForDisconnectedVerts=False, warnIfTooSmall=True):
    if maskShape is None:
      try:
        maskShape = tuple(self.stack().max(0)[::-1]+1)
      except ValueError:
        # Mask is zero-sized
        dtype = 'bool' if asBool else 'uint16'
        return np.zeros((0,0), dtype)
      # Guaranteed not to be too small
      warnIfTooSmall = False
    if warnIfTooSmall:
      cmpShape = maskShape if isinstance(maskShape, Sequence) else maskShape.shape[:2]
      # Wait until inside 'if' so max isn't unnecessarily calculated
      # Edge case: Empty vertices set will trigger a value warning
      if len(self) == 0:
        vertMax = 0
      else:
        vertMax = self.stack().max(0)[::-1]
      if np.any(vertMax > np.array(cmpShape[:2])):
        warn('Vertices don\'t fit in the provided mask size.\n'
             f'Vertex shape: {vertMax}, mask shape: {cmpShape}')
    if checkForDisconnectedVerts:
      fillArg = []
      for verts in self: # type: XYVertices
        if verts.connected:
          fillArg.append(verts)
        else:
          # Make sure each point is treated separately, not part of a shape
          # to fill
          fillArg.extend(verts)
    else:
      fillArg = self
    if isinstance(maskShape, NChanImg):
      out = maskShape
    else:
      out = np.zeros(maskShape, 'uint16')
    nChans = 1 if out.ndim < 3 else out.shape[2]
    if fillColor is None:
      fillColor = tuple([1 for _ in range(nChans)])
    fillColorCmp = np.array(fillColor)
    if np.any((np.iinfo(out.dtype).min > fillColorCmp)
              | (fillColorCmp > np.iinfo(out.dtype).max)):
      raise ValueError('Fill color is larger or smaller than mask range can represent')
    cv.fillPoly(out, fillArg, fillColor)
    if asBool:
      return out > 0
    else:
      return out

  @staticmethod
  def fromBwMask(bwMask: BlackWhiteImg, simplifyVerts=True, externOnly=False) -> ComplexXYVertices:
    approxMethod = cv.CHAIN_APPROX_SIMPLE
    if not simplifyVerts:
      approxMethod = cv.CHAIN_APPROX_NONE
    retrMethod = cv.RETR_CCOMP
    if externOnly:
      retrMethod = cv.RETR_EXTERNAL
    # Contours are on the inside of components, so dilate first to make sure they are on the
    # outside
    #bwmask = dilation(bwmask, np.ones((3,3), dtype=bool))
    if bwMask.dtype != np.uint8:
      bwMask = bwMask.astype('uint8')
    contours, hierarchy = cv.findContours(bwMask, retrMethod, approxMethod)
    compVertices = ComplexXYVertices()
    for contour in contours:
      compVertices.append(XYVertices(contour[:,0,:]))
    if hierarchy is None:
      hierarchy = np.ones((0,1,4), int)*-1
    else:
      hierarchy = hierarchy[0,:,:]
    compVertices.hierarchy = hierarchy
    return compVertices

  def __str__(self) -> str:
    """
    Improve the readability of vertex list in table by just displaying stats of larger arrays
    :return: Human readable string representation
    """
    concatVerts = self.stack()
    if len(concatVerts) <= 4: return str(concatVerts)
    return f'Mean:\t{np.round(concatVerts.mean(0), 1)}\n' \
           f'# Points:\t{len(concatVerts)}\n' \
           f'Min:\t{concatVerts.min(0)}\n' \
           f'Max:\t{concatVerts.max(0)}'

  def __eq__(self, other: ComplexXYVertices):
    if len(self) != len(other):
      return False
    for selfVerts, otherVerts in zip(self, other):
      if not np.array_equal(selfVerts, otherVerts):
        return False
    return True

  def __ne__(self, other):
    return not self == other

  def copy(self) -> ComplexXYVertices:
    return ComplexXYVertices([lst.copy() for lst in self], self.hierarchy)

  def serialize(self):
    return str([arr.tolist() for arr in self])

  @staticmethod
  def deserialize(strObj: str) -> ComplexXYVertices:
    # TODO: Infer appropriate hierarchy from the serialized string. It is possible by finding whether vertices are given
    #  in CW or CCW order. This doesn't affect how they are drawn, but it does effect the return values of "holeVerts()"
    #  and "filledVerts()"
    outerLst = literal_eval(strObj)
    return ComplexXYVertices([XYVertices(lst) for lst in outerLst])


import copy
import sys
from datetime import datetime
from functools import lru_cache
from pathlib import Path
from typing import List, Union, Tuple, Any, Optional, Callable, Dict
from warnings import warn

import numpy as np
from pandas import DataFrame as df
from pyqtgraph.Qt import QtCore
from pyqtgraph.parametertree import Parameter
from ruamel.yaml import YAML
from utilitys import ParamEditor, PrjParam

from s3a.constants import TABLE_DIR, REQD_TBL_FIELDS, PRJ_CONSTS, BASE_DIR
from s3a.structures import PrjParamGroup, FilePath
from s3a.generalutils import hierarchicalUpdate
from utilitys import fns
from utilitys.fns import warnLater

yaml = YAML()

def genParamList(nameIter, paramType, defaultVal, defaultParam='value'):
  """Helper for generating children elements"""
  return [{'name': name, 'type': paramType, defaultParam: defaultVal} for name in nameIter]

def _filterForParam(param: PrjParam):
  """Constructs a filter for the parameter based on its type"""
  children = []
  pType = param.pType.lower()
  paramWithChildren = {'name': param.name, 'type': 'group', 'children': children}
  children.append(dict(name='Active', type='bool', value=False))
  if pType in ['int', 'float']:
    retVal = genParamList(['min', 'max'], pType, 0)
    retVal[0]['value'] = -sys.maxsize
    retVal[1]['value'] = sys.maxsize
    children.extend(retVal)
  elif pType in ['prjparam', 'enum', 'list', 'popuplineeditor', 'bool']:
    if pType == 'prjparam':
      iterGroup = [param.name for param in param.value.group]
    elif pType == 'enum':
      iterGroup = [param for param in param.value]
    elif pType == 'bool':
      iterGroup = [f'{param.name}', f'Not {param.name}']
    else: # pType == 'list' or 'popuplineeditor'
      iterGroup = param.opts['limits']
    optsParam = Parameter.create(name='Options', type='group', children=genParamList(iterGroup, 'bool', True))
    def changeOpts(allowed: bool):
      for param in optsParam.childs:
        param.setValue(allowed)
    paramWithChildren = Parameter.create(**paramWithChildren)
    actions = [Parameter.create(name=name, type='action') for name in ('Select All', 'Clear All')]
    actions[0].sigActivated.connect(lambda: changeOpts(True))
    actions[1].sigActivated.connect(lambda: changeOpts(False))
    paramWithChildren.addChildren(actions)
    paramWithChildren.addChild(optsParam)
  elif 'xyvertices' in pType:
    minMax = _filterForParam(PrjParam('', 5))
    minMax.removeChild(minMax.childs[0])
    minMax = minMax.saveState()['children']
    xyVerts = genParamList(['X Bounds', 'Y Bounds'], 'group', minMax, 'children')
    children.extend(xyVerts)
  elif pType in ['str', 'text']:
    # Assumes string
    children.append(dict(name='Regex Value', type='str', value=''))
  else:
    # Don't know how to handle the parameter
    return None

  if isinstance(paramWithChildren, dict):
    paramWithChildren = Parameter.create(**paramWithChildren)
  return paramWithChildren

def filterParamCol(compDf: df, column: PrjParam, filterOpts: dict):
  # TODO: Each type should probably know how to filter itself. That is,
  #  find some way of keeping this logic from just being an if/else tree...
  pType = column.pType
  # idx 0 = value, 1 = children
  dfAtParam = compDf.loc[:, column]

  if pType in ['int', 'float']:
    curmin, curmax = [filterOpts[name]['value'] for name in ['min', 'max']]

    compDf = compDf.loc[(dfAtParam >= curmin) & (dfAtParam <= curmax)]
  elif pType == 'bool':
    filterOpts = filterOpts['Options']['children']
    allowTrue, allowFalse = [filterOpts[name]['value'] for name in
                             [f'{column.name}', f'Not {column.name}']]

    validList = np.array(dfAtParam, dtype=bool)
    if not allowTrue:
      compDf = compDf.loc[~validList]
    if not allowFalse:
      compDf = compDf.loc[validList]
  elif pType in ['prjparam', 'list', 'popuplineeditor']:
    existingParams = np.array(dfAtParam)
    allowedParams = []
    filterOpts = filterOpts['Options']['children']
    if pType == 'prjparam':
      groupSubParams = [p.name for p in column.value.group]
    else:
      groupSubParams = column.opts['limits']
    for groupSubParam in groupSubParams:
      isAllowed = filterOpts[groupSubParam]['value']
      if isAllowed:
        allowedParams.append(groupSubParam)
    compDf = compDf.loc[np.isin(existingParams, allowedParams)]
  elif pType in ['str', 'text']:
    allowedRegex = filterOpts['Regex Value']['value']
    isCompAllowed = dfAtParam.str.contains(allowedRegex, regex=True, case=False)
    compDf = compDf.loc[isCompAllowed]
  elif pType in ['complexxyvertices', 'xyvertices']:
    vertsAllowed = np.ones(len(dfAtParam), dtype=bool)

    xParam = filterOpts['X Bounds']['children']
    yParam = filterOpts['Y Bounds']['children']
    xmin, xmax, ymin, ymax = [param[val]['value'] for param in (xParam, yParam) for val in ['min', 'max']]

    for vertIdx, verts in enumerate(dfAtParam):
      if pType == 'complexxyvertices':
        stackedVerts = verts.stack()
      else:
        stackedVerts = verts
      xVerts, yVerts = stackedVerts.x, stackedVerts.y
      isAllowed = np.all((xVerts >= xmin) & (xVerts <= xmax)) & \
                  np.all((yVerts >= ymin) & (yVerts <= ymax))
      vertsAllowed[vertIdx] = isAllowed
    compDf = compDf.loc[vertsAllowed]
  else:
    warnLater('No filter type exists for parameters of type ' f'{pType}.'
              f' Did not filter column {column.name}.',
              UserWarning)
  return compDf

class TableFilterEditor(ParamEditor):
  def __init__(self, paramList: List[PrjParam]=None, parent=None):
    if paramList is None:
      paramList = []
    _FILTER_PARAMS = [
      _filterForParam(param) for param in paramList
    ]
    super().__init__(parent, paramList=_FILTER_PARAMS, saveDir=TABLE_DIR,
                     fileType='filter', name='&Component Table Filter')

  def updateParamList(self, paramList: List[PrjParam]):
    newParams = []
    badCols = []
    for param in paramList:
      try:
        curFilter = _filterForParam(param)
      except KeyError:
        curFilter = None
      if curFilter is None:
        badCols.append(param)
      else:
        newParams.append(curFilter)
    self.params.clearChildren()
    self.params.addChildren(newParams)
    if len(badCols) > 0:
      colNames = [f'"{col}"' for col in badCols]
      colTypes = np.unique([f'"{col.pType}"' for col in badCols])
      warnLater(f'The table does not know how to create a filter for fields'
            f' {", ".join(colNames)}'
            f' since types {", ".join(colTypes)} do not have corresponding filters', UserWarning)
    self.applyChanges()

  @property
  def activeFilters(self):
    filters = {}
    for child in self.params.childs:
      if child['Active']:
        cState = child.saveState('user')
        keepChildren = cState['children']
        keepChildren.pop('Active')
        filters[child.name()] = keepChildren
    return filters

  def filterCompDf(self, compDf: df):
    strNames = [str(f) for f in compDf.columns]
    for fieldName, opts in self.activeFilters.items():
      try:
        matchIdx = strNames.index(fieldName)
      except IndexError:
        continue
      col = compDf.columns[matchIdx]
      compDf = filterParamCol(compDf, col, opts)
    return compDf

class TableData(QtCore.QObject):
  sigCfgUpdated = QtCore.Signal(object)
  """dict (self.cfg) during update"""

  def __init__(self):
    super().__init__()
    self._factories: Dict[PrjParam, Callable[[], Any]] = {}

    self.filter = TableFilterEditor()
    self.paramParser: Optional[YamlParser] = None

    self.cfgFname: Optional[Path] = None
    self.cfg: Optional[dict] = None

    self.allFields: List[PrjParam] = []
    self.resetLists()

  def makeCompDf(self, numRows=1) -> df:
    """
    Creates a dataframe for the requested number of components.
    This is the recommended method for component instantiation prior to table insertion.
    """
    df_list = []
    dropRow = False
    if numRows <= 0:
      # Create one row and drop it, which ensures data types are correct in the empty
      # dataframe
      numRows = 1
      dropRow = True
    populators = []
    for f in self.allFields:
      if f in self._factories:
        val = self._factories[f]()
      else:
        val = f.value
      populators.append(val)

    for _ in range(numRows):
      # Make sure to construct a separate component instance for
      # each row no objects have the same reference
      df_list.append(copy.copy(populators))
    outDf = df(df_list, columns=self.allFields).set_index(REQD_TBL_FIELDS.INST_ID, drop=False)
    # Set the metadata for this application run
    outDf[REQD_TBL_FIELDS.SRC_IMG_FILENAME] = PRJ_CONSTS.ANN_CUR_FILE_INDICATOR.value
    if dropRow:
      outDf = outDf.drop(index=REQD_TBL_FIELDS.INST_ID.value)
    return outDf

  def addFieldFactory(self, fieldLbl: PrjParam, factory: Callable[[], Any]):
    """
    For fields that are simple functions (i.e. don't require input from the user), a
    factory can be used to create default values when instantiating new table rows.

    :param fieldLbl: WHich field this factory is used for instead of just the default value
    :param factory: Callable to use instead of field value. This is called with no parameters.
    """
    self._factories[fieldLbl] = factory

  def makeCompSer(self):
    return self.makeCompDf().squeeze()

  def loadCfg(self, cfgFname: FilePath=None, cfgDict: dict=None, force=False):
    """
    Lodas the specified table configuration file for S3A. Alternatively, a name
    and dict pair can be supplied instead.
    :param cfgFname: If *cfgDict* is *None*, this is treated as the file containaing
      a YAML-compatible table configuration dictionary. Otherwise, this is the
      configuration name assiciated with the given dictionary.
    :param cfgDict: If not *None*, this is the config data used instad of
      reading *cfgFname* as a file.
    :param force: If *True*, the new config will be loaded even if it is the same name as the
    current config
    """

    _, baseCfgDict = fns.resolveYamlDict(BASE_DIR/'tablecfg.yml')
    cfgFname, cfgDict = fns.resolveYamlDict(cfgFname, cfgDict)
    cfgFname = cfgFname.resolve()
    if not force and self.cfgFname == cfgFname:
      return None
    hierarchicalUpdate(baseCfgDict, cfgDict)

    cfg = baseCfgDict
    if not force and cfg == self.cfg:
      # No need to update things
      return

    self.cfgFname = cfgFname
    self.cfg = cfg
    self.paramParser = YamlParser(cfg)
    self.resetLists()
    for field in cfg.get('fields', {}):
      param = self.paramParser['fields', field]
      if param in self.allFields:
        continue
      param.group = self.allFields
      self.allFields.append(param)

    self.filter.updateParamList(self.allFields)
    self.sigCfgUpdated.emit(self.cfg)

  def clear(self):
    self.loadCfg(cfgDict={})

  def resetLists(self):
    self.allFields.clear()
    self.allFields.extend(REQD_TBL_FIELDS)

  def fieldFromName(self, name: Union[str, PrjParam]):
    """
    Helper function to retrieve the PrjParam corresponding to the field with this name
    """
    return PrjParamGroup.fieldFromParam(self.allFields, name)

NestedIndexer = Union[str, Tuple[Union[str,int],...]]
class YamlParser:
  def __init__(self, cfg: dict):
    self.cfg = cfg

  def parseParamList(self, listName: NestedIndexer, groupOwner: Union[List, PrjParamGroup]=None):
    """
    A simple list is only a list of strings. A complex list is a dict, where each
    (key, val) pair is a list element. See the structural setup in pg.ListParameter.
    """
    if not isinstance(listName, tuple):
      listName = (listName,)
    paramList = self.getNestedCfgName(listName)
    outList = []
    if groupOwner is None:
      groupOwner = outList
    for ii in range(len(paramList)):
      # Default to list of string values
      accessor = listName + (ii,)
      curParam = self[accessor]
      curParam.group = groupOwner
      outList.append(curParam)
    return outList

  @lru_cache(maxsize=None)
  def __getitem__(self, paramName: NestedIndexer):
    value = self.getNestedCfgName(paramName)
    if not isinstance(paramName, tuple):
      paramName = (paramName,)
    leafName = paramName[-1]
    # Assume leaf until proven otherwise since most mechanics are still applicable
    if not isinstance(value, dict):
      parsedParam = self.parseLeaf(leafName, value)
    else:
      value = value.copy()
      # Format nicely for PrjParam creation
      nameArgs = {'value': value.pop('value', None),
                  'pType': value.pop('pType', None),
                  'helpText': value.pop('helpText', '')}
      # Forward additional args if they exist
      parsedParam = PrjParam(leafName, **nameArgs, **value)
    return parsedParam

  def parseLeaf(self, paramName: str, value: Any):
    leafParam = PrjParam(paramName, value)
    value = leafParam.value
    if isinstance(value, bool):
      pass
      # Keeps 'int' from triggering
    elif isinstance(value, float):
      leafParam.pType = 'float'
    elif isinstance(value, int):
      leafParam.pType = 'int'

    elif isinstance(value, list):
      # If inner values are dicts, it could be a complex list. Otherwise,
      # it is a primitive list
      testVal = value[0]
      if isinstance(testVal, dict):
        leafParam = self.parseParamList(leafParam.name)[0]
      else:
        # List of primitive values
        leafParam = value
    return leafParam

  def getNestedCfgName(self, namePath: NestedIndexer):
    if isinstance(namePath, str):
      namePath = (namePath,)
    out = self.cfg
    while len(namePath) > 0:
      out = out[namePath[0]]
      namePath = namePath[1:]
    return out


from pathlib import Path

from setuptools import setup, find_packages

__version__ = ''
line = open(Path(__file__).parent/f's3a/__version__.py').readline()
exec(line)

def _mltGlob(curDir, *globs):
  curDir = Path(curDir)
  outFiles = []
  for curGlob in globs:
    outFiles.extend(curDir.glob(curGlob))
  outFiles = [str(f) for f in outFiles]
  return outFiles

here = Path(__file__).parent
lines = (here/'requirements.txt').open('r').readlines()
required = []
for line in lines:
  if not line.startswith('#'):
    required.append(line.strip('\n'))

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(
  name='s3a',
  version=__version__,
  package_dir={'s3a':'s3a'},
  packages=find_packages(),
  data_files=[('s3a/s3a/icons', _mltGlob('s3a/icons', '*.svg', '*.png')),
  ('s3a/s3a', _mltGlob('s3a/', '*.yml'))],
  entry_points={
    'console_scripts': [
      's3a-gui = s3a.__main__:main',
    ]
  },

  install_requires=required,
  include_package_data=True,
  url='https://gitlab.com/ficsresearch/s3a',
  download_url='https://gitlab.com/ficsresearch/s3a',
  license='MIT',
  description='Semantic labeling made simple',
  long_description=long_description,
  long_description_content_type='text/markdown',
  keywords='semantic, labeling, image, annotation',
  author='Nathan Jessurun',
  author_email='njessurun@ufl.edu',
  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3',      #Specify which python versions that you want to support
  ],
  python_requires='>=3.7.*',
)
